from .Utils import *
from .Logger import *
import yfinance as yf
import pandas as pd
import json

class DataFetch:
    """DataFetch is a class that handles fetching data from the internet and caching it locally."""

    def __init__(self, cachePath = GetDataPath(), logger = Logger("PPFT")) -> None:
        """Initializes the DataFetch class"""
        self._logger = logger
        self._cachePath = cachePath

    def GetStockInfo(self, tickerName, forceRedownload = False):
        """Gets stock info for a given ticker name"""
        cachedDataLocation = os.path.join(self._cachePath, f'cache/StockInfo/{tickerName}')
        VerifyPath(cachedDataLocation)
        cachedDataPath = os.path.join(cachedDataLocation, f'info.json')
        if PathExists(cachedDataPath) and not forceRedownload:
            self._logger.Log(f'Using cached data for {tickerName} [{cachedDataPath}]')                        
            with open(cachedDataPath, 'r') as f:
                return json.load(f)
        else:
            self._logger.Log(f'Failed to find cached data for {tickerName}')
            self._logger.Log(f'Fetching data for {tickerName} at [{cachedDataPath}]')
            ticker = yf.Ticker(tickerName)
            data = ticker.info
            with open(cachedDataPath, 'w') as f:
                json.dump(data, f)
            return data
        return None
    
    def GetStockHistory(self, tickerName, interval = '1d', start = None, end = None, forceRedownload = False, proxy = None):
        """Gets stock history for a given ticker name"""
        cachedDataLocation = os.path.join(self._cachePath, f'cache/StockInfo/{tickerName}')
        VerifyPath(cachedDataLocation)
        cachedDataPath = os.path.join(cachedDataLocation, f'history-{start}-{end}-{interval}.csv')
        if PathExists(cachedDataPath) and not forceRedownload:
            self._logger.Log(f'Using cached data for {tickerName} [{cachedDataPath}]')                        
            with open(cachedDataPath, 'r') as f:
                return pd.read_csv(f)                
        else:
            self._logger.Log(f'Failed to find cached data for {tickerName}')
            self._logger.Log(f'Fetching data for {tickerName} at [{cachedDataPath}]')
            ticker = yf.Ticker(tickerName)
            data = ticker.history(interval = interval, start = start, end = end, proxy = proxy)
            data.to_csv(cachedDataPath, encoding = 'utf-8')
            return data
        return None
        


import os
import sys
from datetime import datetime

class Logger:
    """A simple logger class"""

    def __init__(self, name, logToConsole = True, path = None):
        """Initializes the logger"""
        self._name = name
        self._path = path
        self._logToConsole = logToConsole

    def Log(self, message):
        """Logs a message in format [name] [timestamp] [message]"""
        message = f'[{self._name}] [{datetime.now().strftime("%Y-%m-%d %H:%M:%S")}] {message}'
        if self._logToConsole:
            print(message)
        if self._path is not None:
            with open(self._path, 'a') as f:
                f.write(message + '\n')
        
        


from .DataFetch import *
from .Portfolio import *
from .PortfolioInputWindow import *

import imgui
import glfw
import OpenGL.GL as gl
from imgui.integrations.glfw import GlfwRenderer

def main():
    # df = DataFetch()
    # print(df.GetStockInfo('GOOG'))
    # print(df.GetStockHistory("GOOG", "1d", "2005-01-01", "2022-12-31"))
    # pf = Portfolio("Test", ["GOOG", "MSFT", "AAPL", "NVDA"])
    # pf.Analyse("2020-01-01", "2022-12-31")

    portfolioInputWindow = PortfolioInputWindow()

    InitializeGLFW()
    window = CreateWindow(500, 500, "Portfolio Tools 2.0 - Jaysmito Mukherjee")
    gl.glClearColor(0.2, 0.2, 0.2, 1.0)
    imgui.create_context()
    SetupImGuiTheme()
    impl = GlfwRenderer(window)
    while not glfw.window_should_close(window):
        window_size = glfw.get_window_size(window)
        gl.glViewport(0, 0, window_size[0], window_size[1])
        gl.glClearColor(0.2, 0.2, 0.2, 1.0)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        
        imgui.new_frame()
        
        portfolioInputWindow.Show()
        
        imgui.end_frame()

        imgui.render()
        impl.render(imgui.get_draw_data())
        
        glfw.swap_buffers(window)
        glfw.poll_events()
        impl.process_inputs()
    
    impl.shutdown()
    glfw.terminate()

if __name__ == "__main__":
    main()


STOCK_TICKER_NAMES = {
    'AAPL': 'Apple Inc.',
    'MSFT': 'Microsoft Corporation',
    'AMZN': 'Amazon.com, Inc.',
    'GOOG': 'Alphabet Inc. (Class C)',
    'TSLA': 'Tesla, Inc.',
    'BABA': 'Alibaba Group Holding Limited (ADR)',
    'NVDA': 'NVIDIA Corporation',
    'JPM': 'JPMorgan Chase & Co.',
    'V': 'Visa Inc.',
    'BAC': 'Bank of America Corporation',
    'WMT': 'Walmart Inc.',
    'MA': 'Mastercard Incorporated',
    'PG': 'The Procter & Gamble Company',
    'JNJ': 'Johnson & Johnson',
    'KO': 'The Coca-Cola Company',
    'PEP': 'PepsiCo, Inc.',
    'DIS': 'The Walt Disney Company',
    'LVMUY': 'LVMH Moët Hennessy - Louis Vuitton, Société Européenne (ADR)',
    'LMT': 'Lockheed Martin Corporation',
    'NKE': 'NIKE, Inc.',
    'HDB': 'HDFC Bank Limited (ADR)',
    'TM': 'Toyota Motor Corporation (ADR)',
    'PTR': 'PetroChina Company Limited (ADR)',
    'RIO': 'Rio Tinto Group',
    'BHP': 'BHP Group',
    'SNY': 'Sanofi',
    'NVO': 'Novo Nordisk A/S (ADR)',
    'EL': 'The Estée Lauder Companies Inc.',
    'LRLCY': 'L\'Oréal S.A. (ADR)',
    'SAN': 'Banco Santander, S.A. (ADR)',
    'ING': 'ING Groep N.V. (ADR)',
    'ICAGY': 'International Consolidated Airlines Group, S.A. (ADR)',
    'BTI': 'British American Tobacco p.l.c. (ADR)',
    'UL': 'Unilever PLC (ADR)',
    'BP': 'BP p.l.c.',
    'TOT': 'TotalEnergies SE',
    'VOD': 'Vodafone Group Plc (ADR)',
    'NESN': 'Nestlé S.A.',
    'UBER': 'Uber Technologies, Inc.',
    'DTEGY': 'Deutsche Telekom AG (ADR)',
    'BMWYY': 'Bayerische Motoren Werke Aktiengesellschaft (ADR)',
    'TCEHY': 'Tencent Holdings Limited (ADR)',
    'NOK': 'Nokia Corporation (ADR)',
    'AZN': 'AstraZeneca PLC (ADR)',
    'SNE': 'Sony Corporation (ADR)',
    '9984.T': 'SoftBank Group Corp.',
    '6502.T': 'Toshiba Corporation',
    '7203.T': 'Toyota Motor Corporation',
    'AAPL34.SA': 'Apple Inc.',
    'VALE3.SA': 'Vale S.A.',
    'B3SA3.SA': 'B3 S.A. - Brasil, Bolsa, Balcão',
    'PETR4.SA': 'Petroleo Brasileiro S.A. - Petrobras'
}



import imgui
from .Utils import *
from .Portfolio import *
from .PortfolioManager import *
from .Tickers import *

class PortfolioInputWindow :
    def __init__(self):
        self._title = "Portfolio Input"
        self._portfolioStocks = []
        self._portfolioName = "Portfolio 0"
        self._portfolios = []
        self._startDate = "2020-01-01"
        self._endDate = "2022-12-31"
    
    def ShowExistingPortfoliosPopup(self):
        """Show a popup modal with a list of existing portfolios to choose from and a button to remove them or open them"""
        with imgui.begin_popup_modal("Existing Portfolios") as select_popup:
            if select_popup.opened:
                for portfolio in self._portfolios:
                    imgui.text(portfolio._portfolio._name)
                    imgui.same_line()
                    if imgui.button("Open"):
                        portfolio._showWindow = True
                        imgui.close_current_popup()
                    imgui.same_line()
                    if imgui.button("Remove"):
                        self._portfolios.remove(portfolio)
                imgui.separator()
                if imgui.button("Close"):
                    imgui.close_current_popup()
            
    def Show(self):
        imgui.begin(self._title)

        if imgui.button("Add Stock"):
            stockItem = Object()
            stockItem.tickerId = 0
            stockItem.quantity = 10
            self._portfolioStocks.append(stockItem)
        
        imgui.same_line()

        if imgui.button("Clear All"):
            self._portfolioStocks.clear()
        
        if imgui.button("Add Portfolio") and len(self._portfolioStocks) > 0:
            totalStocksInPortfolio = sum([stock.quantity for stock in self._portfolioStocks])
            weights = [stock.quantity / totalStocksInPortfolio for stock in self._portfolioStocks]
            stocks = [list(STOCK_TICKER_NAMES.keys())[stock.tickerId] for stock in self._portfolioStocks]
            portfolio = Portfolio(self._portfolioName, stocks, weights, None)
            self._portfolios.append(PortfolioManager(portfolio))
            self._portfolioStocks.clear()
            self._portfolioName = f"Portfolio {len(self._portfolios)}"
        
        imgui.same_line()

        if imgui.button("Select Existing Portfolio"):
            imgui.open_popup("Existing Portfolios")

        self.ShowExistingPortfoliosPopup()

        changed, self._portfolioName = imgui.input_text("Portfolio Name", self._portfolioName)

        for stock in self._portfolioStocks:
            imgui.separator()
            imgui.push_id(str(stock))
            stock.tickerId = ImGuiShowComboBox("Ticker", list(STOCK_TICKER_NAMES.values()), stock.tickerId)
            changed, stock.quantity = imgui.slider_int("Quantity", stock.quantity, 0, 100)
            if imgui.button("Remove"):
                self._portfolioStocks.remove(stock)
                imgui.pop_id()
                break
            imgui.pop_id()

        imgui.separator()       
        imgui.end()

        for portfolio in self._portfolios:
            portfolio.Show()


__all__ = ['Utils', 'Logger', 'DataFetch', 'Portfolio', 'PortfolioInputWindow']

from .Utils import *
from .Logger import *
from .DataFetch import *
from .Predictor import *
from .Portfolio import *
from .PortfolioInputWindow import *



from .Utils import *
from .Logger import *
from .DataFetch import *
from .Portfolio import *
from .StocksAllocator import *
import pandas as pd
import numpy as np
import json
import os
import imgui

class PortfolioManager:

    def __init__(self, portfolio: Portfolio):
        self._portfolio = portfolio
        self._showWindow = True
        self._startDate = "2020-01-01"
        self._endDate = "2022-12-31"
        self._stockHistoryTextures = []
        self._stockHistoryComparisonTexture = None
        self._stocksBasedDistributionPlotTexture = None
        self._countryBasedDistributionPlotTexture = None
        self._sectorBasedDistributionPlotTexture = None
        self._allocator = StocksAllocator()
        self._allocatedWithStocks = None
        self._allocatorTotalBudget = 100000
    
    def Show(self):
        """Show the portfolio in a imgui window"""
        if not self._showWindow:
            return
        
        windowExpanded, self._showWindow = imgui.begin(f"Portfolio {self._portfolio._name}", True)

        with imgui.begin_tab_bar("MainPortfolioTabBar") as tab_bar:
            if tab_bar.opened:
                with imgui.begin_tab_item("General") as item1:
                    if item1.selected:
                        self.ShowGeneralTab()

                with imgui.begin_tab_item("Analysis") as item2:
                    if item2.selected:
                        if not self._portfolio._hasAnalysis:
                            imgui.text("No analysis has been done yet")
                        else:
                            self.ShowAnalysisTab()

        imgui.end()
    
    def ShowGeneralTab(self):
        expanded, visible = imgui.collapsing_header("Portfolio Summary")
        if expanded:
            imgui.text(f"Name: {self._portfolio._name}")
            imgui.text(f"Stocks: ")
            for i in range(len(self._portfolio._stocks)):
                imgui.text(f"    {self._portfolio._stocks[i]}: {self._portfolio._weights[i]}")
            imgui.separator()
        
        expanded, visible = imgui.collapsing_header("Portfolio Analysis")
        if expanded:
            changed, self._startDate = imgui.input_text("Start Date", self._startDate, 10)
            changed, self._endDate = imgui.input_text("End Date", self._endDate, 10)
            if imgui.button("Analyse"):
                self._portfolio.Analyse(self._startDate, self._endDate)
                self.LoadAnalysis()
        
        expanded, visible = imgui.collapsing_header("Portfolio Allocation")
        if expanded and self._portfolio._hasAnalysis:
            if self._allocatedWithStocks == None:
                imgui.text("No allocation has been done yet")
            else:
                imgui.text("Allocated with stocks: ")
                for i in range(len(self._allocatedWithStocks.allocated)):
                    imgui.text(f"    {self._portfolio._stocks[i]} : {self._allocatedWithStocks.allocated[i]} ({self._allocatedWithStocks.prices[i]} each)")
                imgui.text(f"Funds Left: {self._allocatedWithStocks.fundsLeft}")
                imgui.text(f"Funds Used: {self._allocatedWithStocks.totalAllocated}")
            changed, self._allocatorTotalBudget = imgui.input_int("Total Budget", self._allocatorTotalBudget, 1)
            if imgui.button("Allocate Greedy"):
                prices = [self._portfolio._analysis.stockHistory[stock].iloc[-1] for stock in self._portfolio._stocks]
                self._allocatedWithStocks = self._allocator.Allocate(self._portfolio._weights, prices, self._allocatorTotalBudget, "greedy")
            imgui.same_line()
            if imgui.button("Allocate Random"):
                prices = [self._portfolio._analysis.stockHistory[stock].iloc[-1] for stock in self._portfolio._stocks]
                self._allocatedWithStocks = self._allocator.Allocate(self._portfolio._weights, prices, self._allocatorTotalBudget, "random")
        elif expanded:
            imgui.text("No analysis has been done yet")

        
    def ShowAnalysisTab(self):
        expanded, visible = imgui.collapsing_header("Stocks History")
        if expanded:
            imgui.push_id("StocksHistoryTab")
            for i in range(len(self._portfolio._stocks)):
                imgui.text(f"{self._portfolio._stocks[i]}")
                imgui.image(self._stockHistoryTextures[i].id, self._stockHistoryTextures[i].width, self._stockHistoryTextures[i].height)
                if imgui.button(f"Save {self._portfolio._stocks[i]} as File"):
                    SaveImageAsFile(self._stockHistoryTextures[i], ShowFileSaveWindow() + ".png")
                imgui.separator()
            
            imgui.text("Comparison")
            imgui.image(self._stockHistoryComparisonTexture.id, self._stockHistoryComparisonTexture.width, self._stockHistoryComparisonTexture.height)
            if imgui.button("Save Comparison as File"):
                SaveImageAsFile(self._stockHistoryComparisonTexture, ShowFileSaveWindow() + ".png")
            imgui.separator()
            imgui.pop_id()
        
        expanded, visible = imgui.collapsing_header("Portfolio Diversification")
        if expanded:
            imgui.text("Stocks Based Distribution")
            imgui.image(self._stocksBasedDistributionPlotTexture.id, self._stocksBasedDistributionPlotTexture.width, self._stocksBasedDistributionPlotTexture.height)
            if imgui.button("Save Stocks Based Distribution as File"):
                SaveImageAsFile(self._stocksBasedDistributionPlotTexture, ShowFileSaveWindow() + ".png")
            
            imgui.text("Country Based Distribution")
            imgui.image(self._countryBasedDistributionPlotTexture.id, self._countryBasedDistributionPlotTexture.width, self._countryBasedDistributionPlotTexture.height)
            if imgui.button("Save Country Based Distribution as File"):
                SaveImageAsFile(self._countryBasedDistributionPlotTexture, ShowFileSaveWindow() + ".png")
            
            imgui.text("Sector Based Distribution")
            imgui.image(self._sectorBasedDistributionPlotTexture.id, self._sectorBasedDistributionPlotTexture.width, self._sectorBasedDistributionPlotTexture.height)
            if imgui.button("Save Sector Based Distribution as File"):
                SaveImageAsFile(self._sectorBasedDistributionPlotTexture, ShowFileSaveWindow() + ".png")

        expanded, visible = imgui.collapsing_header("Numerical Analysis")
        if expanded:
            imgui.text("Variance: " + str(self._portfolio._analysis.portfolioVariance * 100) + "%")
            imgui.text("Risk: " + str(self._portfolio._analysis.portfolioRisk * 100) + "%")
            imgui.text("Return: " + str(self._portfolio._analysis.portfolioExpectedReturn * 100) + "%")
            imgui.text("Sharpe Ratio: " + str(self._portfolio._analysis.portfolioSharpeRatio))

    def UnloadAnalysis(self):
        for texture in self._stockHistoryTextures:
            DeleteGLTexture(texture.id)

        if self._stockHistoryComparisonTexture != None:
            DeleteGLTexture(self._stockHistoryComparisonTexture.id)
    
        if self._stocksBasedDistributionPlotTexture != None:
            DeleteGLTexture(self._stocksBasedDistributionPlotTexture.id)

        if self._countryBasedDistributionPlotTexture != None:
            DeleteGLTexture(self._countryBasedDistributionPlotTexture.id)
        
        if self._sectorBasedDistributionPlotTexture != None:
            DeleteGLTexture(self._sectorBasedDistributionPlotTexture.id)

    def LoadAnalysis(self):
        self.UnloadAnalysis()
        self._stockHistoryTextures = [ CreateGLTexture(self._portfolio._analysis.stockHistoryPlots[i]) for i in range(len(self._portfolio._stocks)) ]
        self._stockHistoryComparisonTexture = CreateGLTexture(self._portfolio._analysis.stockHistoryComparisonPlot)
        self._stocksBasedDistributionPlotTexture = CreateGLTexture(self._portfolio._analysis.stocksBasedDistributionPlot)
        self._countryBasedDistributionPlotTexture = CreateGLTexture(self._portfolio._analysis.countryBasedDistributionPlot)
        self._sectorBasedDistributionPlotTexture = CreateGLTexture(self._portfolio._analysis.sectorBasedDistributionPlot)


from .Utils import *
from .Logger import *
from .DataFetch import *
import pandas as pd
import numpy as np
import json
import os
import io
import imgui
import matplotlib.pyplot as plt

from pypfopt.efficient_frontier import EfficientFrontier
from pypfopt import risk_models
from pypfopt import expected_returns


class Portfolio:
    """Portfolio is a class that handles the portfolio"""

    def __init__(self, name, stocks, weights = None, predictor = None, dataFetch = DataFetch(), logger = Logger("PPFT")):
        """Initialise the portfolio with the given name, stocks, weights, predictor, dataFetch and logger"""
        self._name = name
        self._stocks = stocks
        if len(stocks) == 0:
            raise RuntimeError("Stocks size cannot be zero")
        if weights == None or len(stocks) != len(weights):
            weights = [1.0 / len(stocks)] * len(stocks)
        self._weights = weights
        self._predictor = predictor
        self._dataFetch = dataFetch
        self._logger = logger
        self._analysis = None
        self._hasAnalysis = False
        self._stockHistory = None
        self._isCompoundingReturns = False
        self._logReturns = False
    
    def FetchStockHistory(self, startDate, endDate, silent = False):
        # is endDate is None, then set it to today
        if endDate == None:
            endDate = str(datetime.now().strftime("%Y-%m-%d"))
        # check if end date after today
        requirePrediction = False
        finalEndDate = endDate
        if datetime.strptime(endDate, "%Y-%m-%d") > datetime.now():
            self._logger.Log(f'End date {endDate} is after today')
            self._logger.Log(f'Using predictor to predict rest of the data')
            requirePrediction = True
            finalEndDate = str(datetime.now().strftime("%Y-%m-%d"))
        # get the stock history till today
        self._logger.Log(f'Loading stock history from {startDate} to {finalEndDate}')
        self._stockHistory = pd.DataFrame()
        for stock in self._stocks:
            self._stockHistory[stock] = self._dataFetch.GetStockHistory(stock, start = startDate, end = finalEndDate)["Close"]
        
        if requirePrediction:
            self._logger.Log(f'Predicting stock history from {finalEndDate} to {endDate}')
            if self._predictor == None:
                raise RuntimeError("Predictor is not set")
            self._stockHistory = self._predictor.Predict(self._stockHistory, finalEndDate, endDate)
        

    
    def NumericalAnalysis(self, startDate, endDate, silent = False):
        """Do a numerical analysis of the portfolio"""

        if not silent:
            self._logger.Log(f"Starging numerical analysis of portfolio {self._name}")
            self._logger.Log(f"Stocks: {self._stocks}")
            self._logger.Log(f"Weights: {self._weights}")
        
        if self._analysis == None:
            self.FetchStockHistory(startDate, endDate, silent)
            self._analysis = Object()
            self._analysis.startDate = startDate
            self._analysis.endDate = endDate
            self._analysis.stockHistory = self._stockHistory.copy()
        
        # calculate the daily returns
        if self._logReturns:
            self._analysis.dailyReturns = np.log(1 + self._stockHistory.pct_change()).dropna(how="all")
        else:
            self._analysis.dailyReturns = self._stockHistory.pct_change().dropna(how="all")
        
        if self._isCompoundingReturns:
            self._analysis.meanDailyReturns = (self._analysis.dailyReturns + 1).prod() ** (252 / self._analysis.dailyReturns.count) - 1
        else:
            self._analysis.meanDailyReturns = self._analysis.dailyReturns.mean() * 252
        # calculate the covariance matrix
        self._analysis.covarianceMatrix = self._analysis.dailyReturns.cov() * 252

        # calculate the portfolio variance
        self._analysis.portfolioVariance = np.dot(self._weights, np.dot(self._analysis.covarianceMatrix, self._weights))

        # calculate the portfolio standard deviation
        self._analysis.portfolioStandardDeviation = np.sqrt(self._analysis.portfolioVariance)
        self._analysis.portfolioRisk = self._analysis.portfolioStandardDeviation

        # calculate the portfolio expected return
        self._analysis.portfolioExpectedReturn = np.dot(self._weights, self._analysis.meanDailyReturns)

        # calculate the portfolio sharpe ratio
        self._analysis.portfolioSharpeRatio = (self._analysis.portfolioExpectedReturn - 0.02) / self._analysis.portfolioStandardDeviation

   

    def Analyse(self, startDate, endDate):
        """Analyse the portfolio from the start date to the end date"""
        self._logger.Log(f'Calculating portfolio history from {startDate} to {endDate}')
        
        if self._analysis == None:
            self.FetchStockHistory(startDate, endDate, False)
            self._analysis = Object()
            self._analysis.startDate = startDate
            self._analysis.endDate = endDate
            self._analysis.stockHistory = self._stockHistory.copy()

        self._analysis.stockHistoryPlots = []
        
        # plot the stock history of individual stocks in individual graphs
        self._logger.Log(f'Plotting stock history of individual stocks')
        for stock in self._stocks:
            plt.plot(self._stockHistory[stock])
            plt.title(stock)
            buffer = io.BytesIO()
            plt.savefig(buffer, format = "png")
            self._analysis.stockHistoryPlots.append(buffer)
            plt.clf()

        # plot the stock history comparison chart
        self._logger.Log(f'Plotting stock history comparison chart')
        normalizedStockHistory = self._stockHistory.divide(self._stockHistory.iloc[0], axis=1)
        plt.plot(normalizedStockHistory)
        plt.title("Stock History Comparison")
        buffer = io.BytesIO()
        plt.savefig(buffer, format = "png")
        self._analysis.stockHistoryComparisonPlot = buffer
        plt.clf()
    
        # plot pie chart of portfolio weights with labels and percentages
        self._logger.Log(f'Plotting Portfolio Distribution')
        plt.pie(self._weights, labels = self._stocks, autopct='%1.1f%%')
        plt.title("Portfolio Distribution")
        buffer = io.BytesIO()
        plt.savefig(buffer, format = "png")
        self._analysis.stocksBasedDistributionPlot = buffer
        plt.clf()

        # load stock infos
        self._logger.Log(f'Loading stock infos')
        self._analysis.stockInfos = {}
        self._analysis.countryBasedDistribution = {}
        self._analysis.sectorBasedDistribution = {}
        for i in range(len(self._stocks)):
            stockInfo = self._dataFetch.GetStockInfo(self._stocks[i])
            self._analysis.stockInfos[self._stocks[i]] = stockInfo
            self._analysis.countryBasedDistribution[stockInfo["country"]] = self._analysis.countryBasedDistribution.get(stockInfo["country"], 0) + self._weights[i]
            self._analysis.sectorBasedDistribution[stockInfo["sector"]] = self._analysis.sectorBasedDistribution.get(stockInfo["sector"], 0) + self._weights[i]
        
        # plot pie chart of portfolio country based distribution with labels and percentages
        self._logger.Log(f'Plotting Country Based Distribution')
        plt.pie(self._analysis.countryBasedDistribution.values(), labels = self._analysis.countryBasedDistribution.keys(), autopct='%1.1f%%')
        plt.title("Country Based Distribution")
        buffer = io.BytesIO()
        plt.savefig(buffer, format = "png")
        self._analysis.countryBasedDistributionPlot = buffer
        plt.clf()

        # plot pie chart of portfolio sector based distribution with labels and percentages
        self._logger.Log(f'Plotting Sector Based Distribution')
        plt.pie(self._analysis.sectorBasedDistribution.values(), labels = self._analysis.sectorBasedDistribution.keys(), autopct='%1.1f%%')
        plt.title("Sector Based Distribution")
        buffer = io.BytesIO()
        plt.savefig(buffer, format = "png")
        self._analysis.sectorBasedDistributionPlot = buffer
        plt.clf()

        
        self.NumericalAnalysis(startDate, endDate, True)

        self._hasAnalysis = True



from setuptools import setup, find_packages
import re

requirements = ['numpy', 'pandas', 'matplotlib', 'yfinance', "PyOpenGL", "glfw", "imgui[glfw]", "Pillow"] 

version = '0.0.6'

if not version:
    raise RuntimeError('version is not set')

readme = 'Coming Soon ...'

setup(
    name = 'py-portfolio-tools',
    author = 'Jaysmito Mukherjee',
    author_email = 'jaysmito101@gmail.com',
    url = 'https://gitlab.com/Jaysmito101/py-portfolio-tools',
    project_urls = {
        'Documentation': 'https://gitlab.com/Jaysmito101/py-portfolio-tools/-/wikis/Home',
        'Issue tracker': 'https://gitlab.com/Jaysmito101/py-portfolio-tools/-/issues',
    },
    version = version,
    packages = find_packages(),
    license = 'MIT',
    description = 'A Python Library and set of GUI tools for Portfolio Management',
    long_description = readme,
    long_description_content_type = 'text/markdown',
    include_package_data = True,
    install_requires = requirements,
    python_requires='>=3.8.0',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Office/Business :: Financial',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'Typing :: Typed',
    ],
)



import numpy as np
import math

from .Utils import *
from .Logger import *


class StocksAllocator:

    def __init__(self, logger = Logger("PPFT")) -> None:
        self._logger = logger

    def _AllocateGreedy(self, weights, prices, total, silent):
        """Allocates the given amount of money to the given stocks using the given weights using a greedy algorithm"""

        if len(weights) != len(prices):
            raise RuntimeError("Weights and prices must be of same size")

        # allocate the money to the stocks
        allocated = [0] * len(weights)
        for i in range(len(weights)):
            allocated[i] = total * weights[i] / prices[i]

        # round the allocations to the nearest integer
        allocated = [int(math.floor(x)) for x in allocated]

        # calculate the total allocated
        totalAllocated = sum([allocated[i] * prices[i] for i in range(len(weights))])
        """
        # if the total allocated is less than the total, then allocate the remaining money to the stock with the highest weight
        if totalAllocated < total:
            remaining = total - totalAllocated
            # move from the stock with the highest weight to lowest weight
            # and check if the remaining money is enough to allocate to the stock
            # allocate the max posible stocks of that stock then move to the next stock

            weightsWithIndex = [(weights[i], i) for i in range(len(weights))]
            weightsWithIndex.sort(key=lambda x: x[0])
            weightsWithIndex.reverse()
            minPrice = min(prices)
            while remaining > minPrice:
                for weight, index in weightsWithIndex:
                    if remaining > prices[index]:
                        allocated[index] += 1
                        remaining -= prices[index]
                    else:
                        break

        totalAllocated = sum([allocated[i] * prices[i] for i in range(len(weights))])
        """

        result =  Object()
        result.allocated = allocated
        result.prices = prices
        result.totalAllocated = totalAllocated
        result.fundsLeft = total - totalAllocated
        return result

    def _AllocateRandom(self, weights, prices, total, silent):
        """Allocates the given amount of money to the given stocks using the given weights using a random algorithm"""
        # randomize the weights
        weights = [np.random.uniform(0, 1) for x in range(len(weights))]
        weights = [x / sum(weights) for x in weights]
        return self._AllocateGreedy(weights, prices, total, silent)

    def Allocate(self, weights, prices, total, method = "greedy", silent = False):
        """Allocates the given amount of money to the given stocks using the given weights"""
        if method == "greedy":
            return self._AllocateGreedy(weights, prices, total, silent)
        elif method == "random":
            return self._AllocateRandom(weights, prices, total, silent)
        else:
            raise RuntimeError(f'Unknown allocation method {method}')





import os
import sys
import OpenGL.GL as gl
import glfw
import imgui
import numpy as np
from PIL import Image
import PIL
from tkinter.filedialog import askopenfilename
from tkinter.filedialog import asksaveasfilename

class Object(object):
    pass

def SetupImGuiTheme():
    """Setup the imgui theme"""
    style = imgui.get_style()
    style.window_padding = (15, 15)
    style.window_rounding = 5.0
    style.frame_padding = (5, 5)
    style.frame_rounding = 4.0
    style.item_spacing = (12, 8)
    style.item_inner_spacing = (8, 6)
    style.indent_spacing = 25.0
    style.scrollbar_size = 15.0
    style.scrollbar_rounding = 9.0
    style.grab_min_size = 5.0
    style.grab_rounding = 3.0
    style.window_border_size = 0
    style.child_border_size = 0
    style.popup_border_size = 0
    style.frame_border_size = 0
    style.colors[imgui.COLOR_TEXT] = (0.80, 0.80, 0.83, 1.00)
    style.colors[imgui.COLOR_TEXT_DISABLED] = (0.24, 0.23, 0.29, 1.00)
    style.colors[imgui.COLOR_WINDOW_BACKGROUND] = (0.06, 0.05, 0.07, 1.00)
    style.colors[imgui.COLOR_POPUP_BACKGROUND] = (0.07, 0.07, 0.09, 1.00)
    style.colors[imgui.COLOR_BORDER] = (0.80, 0.80, 0.83, 0.88)
    style.colors[imgui.COLOR_BORDER_SHADOW] = (0.92, 0.91, 0.88, 0.00)
    style.colors[imgui.COLOR_FRAME_BACKGROUND] = (0.10, 0.09, 0.12, 1.00)
    style.colors[imgui.COLOR_FRAME_BACKGROUND_HOVERED] = (0.24, 0.23, 0.29, 1.00)
    style.colors[imgui.COLOR_FRAME_BACKGROUND_ACTIVE] = (0.56, 0.56, 0.58, 1.00)
    style.colors[imgui.COLOR_TITLE_BACKGROUND] = (0.10, 0.09, 0.12, 1.00)
    style.colors[imgui.COLOR_TITLE_BACKGROUND_COLLAPSED] = (1.00, 0.98, 0.95, 0.75)
    style.colors[imgui.COLOR_TITLE_BACKGROUND_ACTIVE] = (0.07, 0.07, 0.09, 1.00)
    style.colors[imgui.COLOR_SCROLLBAR_BACKGROUND] = (0.10, 0.09, 0.12, 1.00)
    style.colors[imgui.COLOR_SCROLLBAR_GRAB] = (0.80, 0.80, 0.83, 0.31)
    style.colors[imgui.COLOR_SCROLLBAR_GRAB_HOVERED] = (0.56, 0.56, 0.58, 1.00)
    style.colors[imgui.COLOR_SCROLLBAR_GRAB_ACTIVE] = (0.06, 0.05, 0.07, 1.00)
    style.colors[imgui.COLOR_SCROLLBAR_GRAB_ACTIVE] = (0.06, 0.05, 0.07, 1.00)
    style.colors[imgui.COLOR_CHECK_MARK] = (0.80, 0.80, 0.83, 0.31)
    style.colors[imgui.COLOR_SLIDER_GRAB] = (0.80, 0.80, 0.83, 0.31)
    style.colors[imgui.COLOR_SLIDER_GRAB_ACTIVE] = (0.06, 0.05, 0.07, 1.00)
    style.colors[imgui.COLOR_BUTTON] = (0.10, 0.09, 0.12, 1.00)
    style.colors[imgui.COLOR_BUTTON_HOVERED] = (0.24, 0.23, 0.29, 1.00)
    style.colors[imgui.COLOR_BUTTON_ACTIVE] = (0.56, 0.56, 0.58, 1.00)
    style.colors[imgui.COLOR_HEADER] = (0.10, 0.09, 0.12, 1.00)
    style.colors[imgui.COLOR_HEADER_HOVERED] = (0.56, 0.56, 0.58, 1.00)
    style.colors[imgui.COLOR_HEADER_ACTIVE] = (0.06, 0.05, 0.07, 1.00)
    style.colors[imgui.COLOR_RESIZE_GRIP] = (0.00, 0.00, 0.00, 0.00)
    style.colors[imgui.COLOR_RESIZE_GRIP_HOVERED] = (0.56, 0.56, 0.58, 1.00)
    style.colors[imgui.COLOR_RESIZE_GRIP_ACTIVE] = (0.06, 0.05, 0.07, 1.00)
    style.colors[imgui.COLOR_PLOT_LINES] = (0.40, 0.39, 0.38, 0.63)
    style.colors[imgui.COLOR_PLOT_LINES_HOVERED] = (0.25, 1.00, 0.00, 1.00)
    style.colors[imgui.COLOR_PLOT_HISTOGRAM] = (0.40, 0.39, 0.38, 0.63)
    style.colors[imgui.COLOR_PLOT_HISTOGRAM_HOVERED] = (0.25, 1.00, 0.00, 1.00)
    style.colors[imgui.COLOR_TEXT_SELECTED_BACKGROUND] = (0.25, 1.00, 0.00, 0.43)
    style.colors[imgui.COLOR_HEADER] = (0.61, 0.61, 0.62, 0.22)
    style.colors[imgui.COLOR_HEADER_HOVERED] = (0.61, 0.62, 0.62, 0.51)
    style.colors[imgui.COLOR_HEADER_ACTIVE] = (0.61, 0.62, 0.62, 0.83)

def ImGuiShowComboBox(label, items, selected):
    """Shows a combo box"""
    with imgui.begin_combo(label, items[selected]) as combo:
        if combo.opened:
            for i, item in enumerate(items):
                is_selected = (i == selected)
                if imgui.selectable(item, is_selected)[0]:
                    selected = i
                if is_selected:
                    imgui.set_item_default_focus()
    return selected

def CreateGLTexture(something):
    """Creates a GL texture from a file"""
    texture = gl.glGenTextures(1)
    gl.glBindTexture(gl.GL_TEXTURE_2D, texture)
    gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_REPEAT)
    gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_REPEAT)
    gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
    gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
    image = Image.open(something)
    image_data = image.convert("RGBA").tobytes()
    gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGBA, image.width, image.height, 0, gl.GL_RGBA, gl.GL_UNSIGNED_BYTE, image_data)
    result = Object()
    result.id = texture
    result.width = image.width
    result.height = image.height
    result.data = image_data
    return result


def DeleteGLTexture(texture):
    """Deletes a GL texture"""
    gl.glDeleteTextures(1, [texture])

def InitializeGLFW():
    """Initializes GLFW"""
    if not glfw.init():
        raise RuntimeError("Failed to initialize GLFW")
    
    glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 3)
    glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 3)
    glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)
    glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, gl.GL_TRUE)

def ShowFileOpenWindow():
    """Shows a file chooser dialog"""
    return askopenfilename()

def ShowFileSaveWindow():
    """Shows a file chooser dialog"""
    return asksaveasfilename()

def SaveImageAsFile(data, path: str):
    """Saves an image as a file"""
    image = Image.frombytes("RGBA", (data.width, data.height), data.data)
    image.save(path)

def SaveAsFile(data, path: str):
    with open(path, 'wb') as f:
        f.write(data)

def CreateWindow(width : int, height : int, title : str):
    """Creates a GLFW window"""
    window = glfw.create_window(width, height, title, None, None)
    if not window:
        glfw.terminate()
        raise RuntimeError("Failed to create GLFW window")
    glfw.make_context_current(window)
    return window

def PathExists(path):
    """Checks if a path exists"""
    return os.path.exists(path)

def VerifyPath(path):
    """Verifies that a path exists, and if not, creates it"""
    if not os.path.exists(path):
        os.makedirs(path)

def GetDataPath():
    """Returns the user data path for the application"""
    if sys.platform == 'win32':
        return os.path.join(os.environ['APPDATA'], 'py-portfolio-tools')
    elif sys.platform == 'darwin':
        return os.path.join(os.environ['HOME'], 'Library', 'Application Support', 'py-portfolio-tools')
    else:
        return os.path.join(os.environ['HOME'], '.config', 'py-portfolio-tools')